import threading
from time import time
from BTrees.OOBTree import OOBTree

class PTCookItem:
	def __init__(self, file, program, macros):
		self.file = file
		self.program = program
		self.macros = macros
		self.lastAccessedTime = time()

	def touch(self):
		self.lastAccessedTime = time()

class PTCookCache:
	def __init__(self):
		self.lock = threading.Lock()
		self.cache=OOBTree()
		self.cacheStarted=time()

	def addToCache(self, file, program, macros):
		self.lock.acquire()
		try:
			try:
				t = self.cache.items(file)
				if t:
					for t in self.cache[file]:
						self.cache.remove(t)
			except:
				# cache will return key error if we don't have the
				# item, we ignore exceptions and add it.
				pass

			t = PTCookItem(file, program, macros)
			self.cache[file]=t
		finally:
			self.lock.release()

	def getCookedPT(self, file):
		try:
			t = None
			try:
				t = self.cache[file]
			except KeyError:
				return ()

			if t:
				return t.program, t.macros

			else:
				return ()
		finally:
			# We used to release our lock, but, we don't do housekeeping
			# in here any more, so we pass, but leave the structure in place
			# in case we want to do locking on fetch again.
			pass

	def removeCookedPT(self, file):
		self.lock.acquire()
		try:
			try:
				if self.cache[file]:
					del self.cache[file]
			except:
				pass
		finally:
			self.lock.release()

