import os, AccessControl, Acquisition, sys
from Globals import package_home, DevelopmentMode
from zLOG import LOG, ERROR, INFO
from string import join, strip, rstrip, split, lower
from Shared.DC.Scripts.Script import Script, BindingsUI
from Shared.DC.Scripts.Signature import FuncCode
from AccessControl import getSecurityManager
from OFS.Traversable import Traversable
from Products.PageTemplates.Expressions import SecureModuleImporter
from Products.PageTemplates.TALES import Default
from ComputedAttribute import ComputedAttribute
from ExtensionClass import Base
from Acquisition import aq_parent, aq_inner

from tjPageTemplate import tjPageTemplate

from urllib import quote

from tjTALEngine import tjTALEngine
from PTCookCache import PTCookCache

CookedCache = PTCookCache()

class tjPageTemplateFile(Script, tjPageTemplate, Traversable):
	
	meta_type = 'Compiled Page Template (File)'

	func_defaults = None
	func_code = FuncCode((), 0)
	_need__name__=1
	_v_last_read = 0

	_default_bindings = {'name_subpath': 'traverse_subpath'}

	security = AccessControl.ClassSecurityInfo()
	security.declareProtected('View management screens',
							  'read', 'document_src')

	def getCachedCookies(self, filename):
		try:
			return CookedCache.getCookedPT(filename)
		except:
			import traceback
			traceback.print_exc()

	def cacheCookies(self, filename, program, macros):
		try:
			return CookedCache.addToCache(filename, program, macros)
		except:
			import traceback
			traceback.print_exc()
			
	def __init__(self, filename, _prefix=None, **kw):

		self.ZBindings_edit(self._default_bindings)
		if _prefix is None:
			_prefix=SOFTWARE_HOME
		elif type(_prefix) is not type(''):
			_prefix = package_home(_prefix)
		name = kw.get('__name__')
		if name:
			self._need__name__ = 0
			self.__name__ = name
		else:
			self.__name__ = os.path.splitext(os.path.split(filename)[-1])[0]
		if not os.path.splitext(filename)[1]:
			filename = filename + '.zpt'
		self.filename = os.path.join(_prefix, filename)

	def pt_getContext(self):
		root = self.getPhysicalRoot()
		c = {'template': self,
			 'nothing': None,
			 'options': {},
			 'default': Default,
			 'root': root,
			 'request': getattr(root, 'REQUEST', None),
			 'modules': SecureModuleImporter,
			 }
		return c

	def _exec(self, bound_names, args, kw):
		"""Call a Page Template"""
		self._cook_check()

		try:
			self.REQUEST.RESPONSE.setHeader('content-type',
											self.content_type)
		except AttributeError:
			pass

		# Execute the template in a new security context.
		
		security=getSecurityManager()
		bound_names['user'] = security.getUser()
		security.addContext(self)
		try:
			return self._v_program._exec(bound_names, args, kw)
		finally:
			security.removeContext(self)
	

	def pt_macros(self):
		self._cook_check()
		return self._v_macros

	def _cook_check(self):
		try:
			if self._v_last_read and not DevelopmentMode:
				return
				
			__traceback_info__ = self.filename
			try:
				mtime=os.stat(self.filename)[8]
			except:
				mtime=0
			if hasattr(self, '_v_program') and mtime == self._v_last_read:
				return

			try:
				cacheItems = self.getCachedCookies(self.filename)
				if cacheItems:
					self._v_last_read = mtime
					self._v_program, macros = cacheItems
					self._v_macros={}
					for k,v in macros.items():
						self._v_macros[k]=v.__of__(self)
					self._v_program = self._v_program.__of__(self)
					return
			except:
				import traceback
				traceback.print_exc()
				pass

			self.pt_edit(open(self.filename), None)
			self._cook(1)
			if self._v_errors:
				LOG('PageTemplateFile', ERROR, 'Error in template',
					join(self._v_errors, '\n'))
				return
			self._v_last_read = mtime
		finally:
			pass

	def _cook(self, authorized = 0):
		"""Compile the TAL and METAL statments.

		A Page Template must always be cooked, and cooking must not
		fail due to user input.		
		"""
		if not authorized:
			return
		try:
			parser = tjTALEngine(self._text, self.filename)
			self._v_program, macros = parser.compile(self.id)
			self._v_macros={}
			for k,v in macros.items():
				self._v_macros[k]=v.__of__(self)
			try:
				self.cacheCookies(self.filename, self._v_program, macros)
			except:
				import traceback
				traceback.print_exc()
				pass

			self._v_program = self._v_program.__of__(self)
		except:
			import traceback
			traceback.print_exc()
			self._v_errors = ["Compilation failed",
							  "%s: %s" % (sys.exc_info()[:2])]

		self._v_warnings = ()

	def document_src(self, REQUEST=None, RESPONSE=None):
		"""Return expanded document source."""

		if RESPONSE is not None:
			RESPONSE.setHeader('Content-Type', 'text/plain')
		return self.read()

	def _get__roles__(self):
		imp = getattr(aq_parent(aq_inner(self)),
					  '%s__roles__' % self.__name__)
		if hasattr(imp, '__of__'):
			return imp.__of__(self)
		return imp

	__roles__ = ComputedAttribute(_get__roles__, 1)

	def __setstate__(self, state):
		raise StorageError, ("Instance of AntiPersistent class %s "
							 "cannot be stored." % self.__class__.__name__)
