from types import *
from cgi import escape
import string

from cStringIO import StringIO

#from DocumentTemplate.DT_Util import ustr
from ExtensionClass import Base
from OFS.Traversable import Traversable
from Shared.DC.Scripts.Script import Script
from Products.PageTemplates.Expressions import SecureModuleImporter
from Products.PageTemplates.TALES import Default
from ComputedAttribute import ComputedAttribute
from AccessControl import getSecurityManager

from uuidgen import uuidgen

import Expressions
import Acquisition

from tjTALEnginePieces import *

BLOCK_CLOSING_TAG_MAP = {
    "tr": ("tr", "td", "th"),
    "td": ("td", "th"),
    "th": ("td", "th"),
    "li": ("li",),
    "dd": ("dd", "dt"),
    "dt": ("dd", "dt"),
    }

def returnMacroNoRender(ob, ns):
	return ob

class NoSlotError(Exception):
	pass

class ParseError(Exception):
	pass


ustr = str
	
# Mini Script
#class MacroContext(Script, Base, Traversable):
class MacroContext(Script, Base):

	_default_bindings = {'name_subpath': 'traverse_subpath'}	
	
	def __init__(self, name, macro, code='', isMacro=0):
		self.__name__=name
		self.macro = macro
		self.code = code
		self.lookupTable = None
		self.isMacro=isMacro		

	def setLookupTable(self, lookupTable):
		self.lookupTable = lookupTable
		self.isStructure = self.lookupTable.isStructure
		self.lGet = self.lookupTable.get

	def getLoopExpr(self, id):
		return self.lGet(id)

	def getBooleanExpr(self, id, context):
		# expr =  self.lGet(id)
		# return expr(context)
		return self.getRawExpr(id, context)

	def execMacro(self, id, context):
		macroId = self.lGet(id)
		macro = macroId._eval(context, render = returnMacroNoRender).__of__(self)

		return macro(current_context=context)

	def execSlot(self, id, context):
		if self.aq_parent.pt_macros().has_key(id):
			slot = self.aq_parent.pt_macros()[id].__of__(self.aq_parent)
			return slot(current_context=context)
		else:
			raise NoSlotError, id

	def getRawExpr(self, id, context=None):
		return self.lGet(id)(context)

	def getExpr(self, id, context=None):
		def escape(s):
			"""Replace special characters '&', '<' and '>' by SGML entities."""
			return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
		
		returnVal = self.getRawExpr(id, context)
		if returnVal is Default or returnVal is None:
			return returnVal
		
		if isinstance(returnVal, basestring):
			if self.isStructure(id):
				return returnVal
			return escape(returnVal)
		return returnVal

	def getAExpr(self, id, name, context, booleanAttr = False):
		def escape(s):
			"""Replace special characters '&', '<' and '>' by SGML entities."""
			return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace('"', "&quote;")
		
		retVal = ''
		val = self.getExpr(id, context)
		if val and (val is not Default):
			if not booleanAttr:
				retVal = ' '+name+'="'+escape(ustr(val))+'"'
			else:
				retVal = ' '+name
		return retVal

	def getCExpr(self, id, context=None):
		def escape(s):
			"""Replace special characters '&', '<' and '>' by SGML entities."""
			return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
		
		returnVal = self.getRawExpr(id, context)
		if not returnVal or returnVal is Default:
			return ''

		if self.isStructure(id):
			return returnVal
		return escape(ustr(returnVal))

	def getMacroExpr(self, id):
		return self.lGet(id)

	__roles__=('Anonymous',)

	def _exec(self, bound_names, args, kw):
		"""Call a Page Template"""

##		if not kw.has_key('args'):
##			kw['args'] = args
##		bound_names['options'] = kw
		try:
			security=getSecurityManager()
			bound_names['user'] = security.getUser()
		except:
			bound_names['user'] = 'Nobody'
			
		return self.pt_render(extra_context=bound_names)

	def pt_render(self, source=0, extra_context={}, current_context=None):
		# StringIO faster than list.append for small chunks...
		output = StringIO()
		c = self.pt_getContext()
		c.update(extra_context)
		self.execute(output, c, current_context)
		return output.getvalue()

	def pt_getContext(self):
		user = ''
		try:
			security=getSecurityManager()
			user = security.getUser()
		except:
			user = 'Nobody'
			
		c = {'template': self.aq_parent,
			 'options': {},
			 'default': Default,
			 'nothing': None,
			 'request': getattr(self.aq_parent,'REQUEST',None),
			 'user': user,
			 'modules': SecureModuleImporter,
			 }		
		parent = getattr(self, 'aq_parent', None)
		if parent is not None:
			# c['here']=getattr(parent, 'aq_parent', parent)
			c['here'] = self.aq_parent.aq_parent
			if self.isMacro:
				c['template']=parent
			c['container'] = self.aq_inner.aq_parent

			pSelf = self
			ppSelf = self
			while parent is not None:
				ppSelf = pSelf
				pSelf = parent
				parent = getattr(pSelf, 'aq_parent', None)
			c['root'] = ppSelf
		return c
	
	def execute(self, output, context, current_context=None):
		gx = {}
		user = ''
		try:
			security=getSecurityManager()
			user = security.getUser()
		except:
			user = 'Nobody'

		# It's pretty close between using .append vs StringIO
		# larger templates seem to benefit more from using .append
		# --akm

		# lines = []
		out = StringIO()
		
		gx['user']= user
		# gx['pt_write'] = lines.append
		gx['pt_write'] = out.write
		gx['getCExpr'] = self.getCExpr
		gx['getAExpr'] = self.getAExpr
		gx['getRawExpr'] = self.getRawExpr
		gx['execSlot'] = self.execSlot
		gx['getLoopExpr'] = self.getLoopExpr		
		gx['execMacro'] = self.execMacro
		gx['getBooleanExpr'] = self.getBooleanExpr
		gx['NoSlotError'] = NoSlotError
		gx['default'] = Default
		gx.update(self.pt_macros())

		econtext = Expressions.getEngine().getContext(context)		
		if current_context:
			econtext.contexts['loop'] = current_context.contexts['loop']
			econtext.global_vars.update(current_context.global_vars)
			econtext.local_vars.update(current_context.local_vars)
			econtext.repeat_vars=current_context.repeat_vars.copy()

		gx['context']=econtext
		eval(self.macro,gx)
		# output.write(''.join(map(str,lines)))
		output.write(out.getvalue())
		out.close()

	def __call__(self, *args, **kwargs):
		current_context=None
		if not kwargs.has_key('args'):
			kwargs['args'] = args
		if kwargs.has_key('current_context') and kwargs['current_context']:
			current_context = kwargs['current_context']
			
		return self.pt_render(extra_context={'options': kwargs,},
							  current_context=current_context)
		
class tjTALEngine(object):
	__slots__ = ('buffer', 'line', 'offset', 'code', 'blocks', 'lookupTable',
				 'id', 'xcode','page')
	def __init__(self, buffer, id = ''):
		self.buffer = buffer
		self.line = 0
		self.offset = 0
		self.code=[]
		self.blocks=[]
		self.lookupTable=ExpressionLookupTable()
		self.id = id
		self.xcode = None
		self.page = None

	def processText(self, buffer, buffLen):
		tagList = []
		tagString = ''
		c = buffer[self.offset]
		startLine = self.line
		while self.offset < buffLen and c != '<':
			if c == '\n':
				self.line = self.line + 1
			tagList.append(c)
			self.offset = self.offset+1
			if self.offset < buffLen:
				c = buffer[self.offset]

		tagString = ''.join(tagList)
		if tagString:
			self.code.append(Chunk(tagString, startLine, self.line))
	
	def processCommentTag(self, buffer, buffLen):
		tagString = ''
		c = buffer[self.offset]
		ended = 0
		startLine = self.line
		while not ended:
			while self.offset < buffLen and c!= '>':
				if c == '\n':
					self.line = self.line + 1
				tagString = tagString + c
				self.offset = self.offset+1
				if self.offset < buffLen:
					c = buffer[self.offset]
			if c != '>':
				raise ParseError, (self.line, self.offset)

			tagString = tagString + c
			self.offset = self.offset + 1
			if tagString[-3:] == '-->':
				ended = 1
			c = buffer[self.offset]

		self.code.append(Chunk(tagString, startLine, self.line))

	def isCommentTag(self, buffer):
		if self.buffer[self.offset:self.offset+4]=='<!--':
			return 1
		return 0
			
	def processTag(self, buffer, buffLen):
		tagString = ''

		if self.isCommentTag(buffer):
			return self.processCommentTag(buffer, buffLen)
		
		c = buffer[self.offset]

		startLine = self.line
		inQuote = 0
		while self.offset < buffLen:
			if not inQuote and c == '>':
				break
			
			if c == '\n':
				self.line = self.line + 1
				c=' '
			if c == '"':
				inQuote = 1 - inQuote
		
			tagString = tagString + c
			self.offset = self.offset+1
			if self.offset < buffLen:
				c = buffer[self.offset]

		if c != '>':
			raise ParseError, ("Malformed Tag", self.line, self.offset)

		if inQuote:
			raise ParseError, ("Unbalanced Quotes", self.line, self.offset)

		tagString = tagString + c
		self.offset = self.offset + 1
		if tagString[:2]=='<!':
			self.code.append(Chunk(tagString, startLine, self.line))
			return

		self.code.append(Tag(tagString, startLine, self.line))
	
	def parseBuffer(self, buffer):
		buffLen = len(buffer)
		self.offset = 0
		while self.offset < buffLen:
			c = buffer[self.offset]
			if c == '<':
				self.processTag(buffer, buffLen)
			else:
				self.processText(buffer, buffLen)

	# Turn linear list of blocks, into hierarchical tree of blocks
	# With Each HTML Container containing it's blocks as subBlocks
	def blockCode(self):
		self.page = Page()
		currentBlock = self.page
		for item in self.code:
			item.parentBlock=currentBlock
			if item.isTag and item.tagName[0]=='/':
				if not currentBlock.tagName == item.tagName[1:]:
					if BLOCK_CLOSING_TAG_MAP.has_key(item.tagName[1:]):
						closeouts = BLOCK_CLOSING_TAG_MAP[item.tagName[1:]]
						while hasattr(currentBlock, 'parentBlock'):
							if currentBlock.tagName in closeouts:
								currentBlock=currentBlock.parentBlock
							else:
								raise "Parser Error", (self.id, "Missing End Tag %s, Lines: %d - %d"%(currentBlock.tagName, item.startLine, item.endLine))
					else:
						raise "Parser Error", (self.id, "Missing End Tag %s, Lines: %d - %d"%(currentBlock.tagName, item.startLine, item.endLine))
				currentBlock = currentBlock.parentBlock
				continue
			elif not item.isTag:
				currentBlock.addBlock(item)			
				continue
			elif not item.isContainer:
				currentBlock.addBlock(item)							
				continue
			
			currentBlock.addBlock(item)
			currentBlock=item
	
	def compile(self, id=''):
		self.parseBuffer(self.buffer)

		self.blockCode()
		try:
			self.page.expandExpressions(self.lookupTable)
		except:
			print self.id
			raise

		self.page.collapseSelf()
		macros = []
		
		for name, block in self.lookupTable.macros.items():
			code = block.spewCode()
			try:
				xcode = compile(code, name, 'exec')
				macros.append(MacroContext(name, xcode, code, 1))
			except:
				print code
				raise
			
		for name, slot in self.lookupTable.slots.items():
			code = slot.spewCode()
			
			try:
				xcode = compile(code, name, 'exec')
				macros.append(MacroContext(name, xcode, code, 1))
			except:
				print code
				raise

		code = self.page.spewCode()
		
		xcode = compile(code, 'tjTemplate_%s'%(id), 'exec')
		program = MacroContext('main', xcode, code)
		program.setLookupTable(self.lookupTable)
		macroDict={}
		for macro in macros:
			macro.setLookupTable(self.lookupTable)
			macroDict[macro.__name__]=macro
		return program, macroDict

