# (C) Copyright 2004-2006 Andrew K. Milton <akm@unyx.net>
#
# Code to generate unique ids (without forking).

import sys
import sets
import time
import socket
import random

NativeUUIDGen = False

# If your platform has a native uuidgen add it here...
if sys.platform == 'win32':
	try:
		from pythoncom import CreateGuid
		def uuidgen():
			x = CreateGuid()
			return str(x)[1:-1]
		NativeUUIGen = True
	except:
		pass

# Fake it out..
if not NativeUUIDGen:

	UUID_NODE_LEN = 6;

	class UUIDInternalTime(object):
		__slots__ = ('low', 'mid', 'high', 'll')
		def __init__(self, low, mid, high):
			self.low = low
			self.mid = mid
			self.high = high

			self.ll = self.low | (self.mid << 32) | (self.high << 48)

		def setLL(self, ll):
			self.ll = ll
			self.low = (self.ll & 0xFFFFFFFFL)
			self.mid = (self.ll >> 32L) & 0xFFFFL
			self.high = ((self.ll >> 48L) & 0xFFFL ) | ( 1L << 12)

	class UUID(object):
		__slots__ = ('time', 'seq', 'node')
		def __init__(self, timeLow = 0, timeMid = 0, timeHigh = 0,
					 seq = 0, node = [0,] * UUID_NODE_LEN):
			self.time = UUIDInternalTime(timeLow, timeMid, timeHigh)
			self.seq = seq
			self.node = node

		def __str__(self):
			_node = ''.join(map(lambda n:'%02x'%(n), self.node))
			return "%08x-%04x-%04x-%04x-%s"%(
				self.time.low , self.time.mid, self.time.high,
				self.seq, _node)

	uuid_last = UUID(0, 0, 0, 0, [0,] * UUID_NODE_LEN)

	def uuidNode():
		nodes = [None,] * UUID_NODE_LEN

		# It's unlikely we can get a MAC address, so try to grab the
		# IPv4 address (4 bytes) and generate 2 more random bytes
		try:
			host, aliases, addresses = socket.gethostbyaddr(socket.gethostname())
			addr = addresses[0]
			dotted = addr.split('.')
			nodes = map(int, dotted)
			rands = [int(random.random() * 255.0),
					 int(random.random() * 255.0) ]
			nodes = nodes + rands
		# If the above fails, generate a random multicast address
		except:
			for i in range(len(nodes)):
				nodes[i] = int(random.random() * 255.0)

		return nodes

	def uuidTime():
		uTime = 0x01B21DD213814000L
		uTime += int(time.time() * 10000000.0)

		return (uTime & ((1L << 60) - 1L))


	def uuidgen():
		global uuid_last

		uuid = UUID()
		uuid.node = uuidNode()
		xTime = uuidTime()

		if ( uuid_last.time.ll == 0L or
			 uuid_last.node[:4] != uuid.node[:4] ):
			uuid.seq = int(random.random() * 65536.0) & 0x3FFF
		elif uuid_last.time.ll >= xTime:
			uuid.seq = (uuid_last.seq + 1) & 0x3FFF
		else:
			uuid.seq = uuid_last.seq


		uuid_last = uuid
		uuid_last.time.setLL((xTime) & (( 1L << 60) - 1L))

		uuid.seq = (uuid.seq | 0x8000)
		uuid.time.setLL(xTime)

		return str(uuid)


if __name__ == '__main__':
	print uuidgen()
